%**********************************************************
%
%   basics.sl contains functions grouped as follows:
%
%       sorting
%       occurrences
%       list buffers
%       minibuffer
%       compile
%       recent files
%
%**********************************************************

  _debug_info = 1;
  _traceback = 1;

variable b_tmpbf = Null_String;
variable RecentzBuf = "/recentz/";
autoload("ed_delete_word","sue");


%**********************************************************
%   sorting functions
%   sort1 is based on rectangles
%   sort2 is based on character streams
%**********************************************************


define sort1(casE)
%----------------------------------------------------------
%   This sorting function needs a rectangle.
%   The sort key is fixed length across the entire rectangle.
%   The rectangle can be just one column.
%   This sort is case insensitive when called from menu
%   or when casE is 0. Otherwise from the keyboard:
%----------------------------------------------------------
{
    if (is_readonly) return;

    !if (check_rectangle)
    {
        flush ("Draw a rectangle to define sort keys.");
        return;
    }

    toggle_block (0);

    variable
    thisl, thisc,
    begl, endl, begc, endc, n, index, i, keys, sortedz, tmpstr;

    %   global variables N_ROW etc. provided by check_rectangle
    endc = E_COL;
    begc = W_COL;
    endl = S_ROW;
    begl = N_ROW;

    endc++;
    %   sue's rectangles include cursor position
    %   therefore to use a sort key matching rect width
    %   we need to increase right column by 1
    %   this will also allow sorting on a single column

    n = endl - begl + 1;
    %   count lines in rectangle

    keys    = String_Type [n];
    sortedz = String_Type [n];
    %   last element of keys[] takes terminator

    %   extract keys and fill them into keys array

    thisl = what_line();
    thisc = what_column();
    goto_line (begl);
    for (i = 0; i < n; i++)
    {
        goto_column (begc);
        push_mark ();
        goto_column (endc);

        %   pop_mark (0) is implicit in bufsubstr

        keys[i] = bufsubstr ();

        go_down (1);
    }

    if (casE) index = array_sort (keys, "strcmp");
    else index = array_sort (keys, "strcmpi");

    %   get the lines into sortedz array in proper order

    for (i = 0; i < n; i++)
    {
        goto_line (index[i] + begl);
        push_mark ();
        eol ();
        tmpstr = bufsubstr ();
        sortedz[i] = tmpstr;
    }

    %   replace unsorted lines with sorted lines

    goto_line (begl);
    for (i = 0; i < n; i++)
    {
        del_to_eol ();
        insert (sortedz[i]);
        go_down (1);
    }

    %   position cursor at start position
    %   so rectangle can be restored

    goto_line (thisl);
    goto_column (thisc);
    restore_rectangle ();

    if(casE) flush ("case SENSITIVE sort");
    else flush ("case insensitive sort");
}


define sort2 ()
%----------------------------------------------------------
%   This sorting function needs a character stream.
%   The sort key is the string from start column to eol,
%   its length may vary from line to line.
%   For fixed length sort keys, use rectangles.
%   Case insensitive.
%----------------------------------------------------------
{
    variable  i, n, begl, endl, begc, endc, index, keys;

    !if (check_region (0)) return;
    %   region in canonical form

    endl = what_line ();
    endc = what_column ();
    pop_mark (1);
    begl = what_line ();
    begc = what_column ();

    n = endl - begl + 1;
    %   count lines in region

    %   do not end the stream left of start column
    if (endc < begc)
    {
        endc;
        endc = begc;
        begc = ();
    }

    keys = String_Type [n];

    %   extract keys and fill array

    goto_line (begl);
    for (i = 0; i < n; ++i)
    {
        goto_column (begc);
        push_mark ();
        eol ();

        keys[i]  = bufsubstr();

        go_down (1);
    }

    index = array_sort (keys, "strcmpi");

    %  arrange region

    goto_line (endl);
    !if (down (1))
    {
        eol ();
        newline ();
    }

    push_spot ();

    for (i = 0; i < n; ++i)
    {
        goto_line (index[i] + begl);
        line_as_string ();
        %  on stack, we are at eol now
        pop_spot ();
        bol ();
        insert(());
        newline ();
        push_spot ();
    }

    pop_spot ();

    goto_line (begl);
    bol ();
    push_mark ();
    goto_line (endl + 1);
    bol ();
    del_region ();
}



%**********************************************************
%
%   Occurrences
%
%   Finds all occurrences of a keyword in a buffer
%   listing them, as whole lines, in a separate buffer
%   and allowing fast switch from the list to the source.
%
%   Essentially same are the list of C or sl functions
%   and the list of shiftkey functions.
%
%**********************************************************


variable Reference_Buffer = Null_String;


%   PROTOTYPES  -------------------------------------------

define occur ();
define occur_rx ();
define occur_c_format_list ();
define occur_c_functions ();
define occur_format ();
define occur_goto_line ();
define occur_init ();
define occur_quit ();


define occur_down ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves down cursor
%----------------------------------------------------------
{
    !if(down(1)) return;
    bol ();
}


define occur_up ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves up cursor
%----------------------------------------------------------
{
    !if(up(1)) return;
    bol ();
}


define occur_right ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves cursor half-screen right in list of occurrences
%   cursor is actually invisible
%   but repeating command may scroll long lines
%----------------------------------------------------------
{
    variable cc=what_column();
    variable ww=window_info('w');
    goto_column_best_try(cc + 1 + ww/2);
}


define occur_left ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves cursor half-screen left in list of occurrences
%   cursor is actually invisible
%   but repeating command may scroll long lines
%----------------------------------------------------------
{
    variable cc=what_column();
    variable ww=window_info('w');
    goto_column_best_try(cc - 1 - ww/2);
}


define occur_pagedown ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   will move down cursor one page
%----------------------------------------------------------
{
    variable start_line = window_line ();
    % cursor is on start_line

    goto_bottom_of_window ();
    % move to bottom of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        % already at bottom of window, now paginate
        call ("page_down");
        goto_bottom_of_window ();
    }
    bol();
}


define occur_pageup ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   will move up cursor one page
%----------------------------------------------------------
{
    variable start_line = window_line ();
    %   cursor is on start_line

    goto_top_of_window ();
    % move to top of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        %   already top, now paginate
        call ("page_up");
        goto_top_of_window ();
    }
    bol();
}


define occur_help ()
%----------------------------------------------------------
%   ancillary function to occur ()
%----------------------------------------------------------
{
    variable
        stringA="Occur Buffer",
        stringB="    [#]",
        stringC="&Quick Help: Occurrences";

    display_hlp2 (stringA,stringB,stringC);
}


define occur_init ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   will initialize keymap for list of occurrences
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "occur_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<127; i++)
        undefinekey (char(i), $1);

    definekey ("occur_down",            Key_Down, $1);
    definekey ("occur_end",             Key_End, $1);
    definekey ("occur_find",            "^F", $1);
    definekey ("occur_find_again",      Key_F3, $1);
    definekey ("occur_goto_line",       Key_Enter, $1);
    definekey ("occur_help",            Key_F1, $1);
    definekey ("occur_home",            Key_Home, $1);
    definekey ("occur_left",            Key_Left, $1);
    definekey ("occur_pagedown",        Key_PgDn, $1);
    definekey ("occur_pageup",          Key_PgUp, $1);
    definekey ("occur_right",           Key_Right, $1);
    definekey ("occur_save",            "^S", $1);
    definekey ("occur_up",              Key_Up, $1);
    definekey ("occur_test",            "^t", $1);
}


define occur_save ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   saves list of occurrences to disk file
%----------------------------------------------------------
{
    variable f_name = Null_String;
    variable b_name =
    read_path_from_mini ("Save to file:");

    (, f_name) = parse_filename (b_name);
    if (f_name == "") return;

    %   if b_name in use, refuse
    if (check_buffer_filename (b_name))
    {
        flush
        ("Choose different name, file is already associated to a buffer.");
        return;
    }

    %   check file status before proceeding

    variable b_status = file_status (b_name);
    %    2   directory
    %    1   exist
    %    0   not exist
    %   -1   not accessible
    %   -2   invalid path
    %   -3   unknown

    if (b_status == 2);

    else if (b_status == -1)
        error ("File exists and is not accessible");

    else if (b_status == -2)
        error ("Invalid path");

    else if (b_status == -3)
        error ("Undetermined error");

    else if (b_status == 0)
        () = write_buffer (b_name);

    else if (b_status == 1)
    {
        variable ch = -1;
        ch = get_mini_response
        ("File exists: append/overwrite/cancel? (a/o/c) ");

        if (ch==-1) return;

        if (ch=='a' or ch=='A')
        {
            bob ();
            push_mark ();
            eob ();
            append_region_to_file (b_name);
        }

        else if (ch=='o' or ch=='O')
            () = write_buffer (b_name);
    }
    delbuf(whatbuf());
    () = find_file (b_name);
}


define occur_quit (str)
%----------------------------------------------------------
%   ancillary function to occur ()
%   delete buffer with list of occurrences
%----------------------------------------------------------
{
    sw2buf (str);
    set_buffer_modified_flag (0);
    Reference_Buffer=Null_String;
    delbuf (str);
}


define occur_goto_line ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   go to line (as prompted for) in list of occurrences
%----------------------------------------------------------
{
    if (what_line () < 3) return;
    bol ();
    !if (ffind (": ")) return;

    variable listBuf = whatbuf ();
    %   store it to delete it

    Reference_Buffer = substr(whatbuf (),5,-1);
    %   necessary to handle multiple occur buffers at same time
    %   goto_line would jump to line in wrong buffer

    !if (strlen(Reference_Buffer)) return;

    !if (bufferp (Reference_Buffer))
    {
        Reference_Buffer = Null_String;
        variable tmpbf = whatbuf ();
        set_buffer_modified_flag (0);
        sw2buf ("*scratch*");
        delbuf (tmpbf);
        flush ("Reference buffer for this list not found.");
        return;
    }

    bol ();
    push_mark ();

    !if (ffind (": ")) return;
    %   cannot happen, used to delimit region

    variable line = integer (bufsubstr);
    %   mark popped here

    sw2buf (Reference_Buffer);
    goto_line (line);

    delbuf(listBuf);
}


define occur_find ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   finds a string in list of occurrences
%----------------------------------------------------------
{
    variable sstr=Null_String;
    sstr = read_mini2 ("Search for:", "", "");
    !if (strlen (sstr)) return;
    LAST_SEARCH=sstr;

    () = fsearch(sstr);

    %   the line and not the string is highlighted
    %   the string may occur more than once in same line
}


define occur_find_again ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   repeats last search in list of occurrences
%----------------------------------------------------------
{
    %   the line and not the string is highlighted
    %   the string may occur more than once in same line

    eol();
    ()=fsearch(LAST_SEARCH);
}


define occur_format (str)
%----------------------------------------------------------
%   ancillary function to occur ()
%   stub for list of occurrences
%----------------------------------------------------------
{
    sw2buf (o_ListBuffer);
    erase_buffer ();
    newline ();
    loop (8) insert_char (' ');
    insert
    ("Press Enter on a line to go to same line in reference buffer.");
    newline ();
    newline ();
}


define occur_rx (str)
%----------------------------------------------------------
%   occurrences for regex expressions, case sensitive here
%----------------------------------------------------------
{
    push_spot ();

    Reference_Buffer = whatbuf ();
    %   not a special buffer

    %   give the list a name dependent on source
    o_ListBuffer =  "/O/ " + whatbuf ();

    %   get rid of any previous listing when refreshing
    if (bufferp (o_ListBuffer)) delbuf (o_ListBuffer);

    occur_format (str);

    sw2buf (Reference_Buffer);

    bob ();
    variable ln=0;

    LAST_SEARCH = str;
    CASE_SEARCH = 1;

    while (re_fsearch (str))
    {
        %   entire line to the stack
        line_as_string ();

        ln = what_line ();

        setbuf (o_ListBuffer);
        if (ln > 99999)
        {
            flush("Line number too big, exiting ...");
            break;
        }
        else vinsert ("%5d: ", ln);

        %   insert from the stack
        insert (());
        newline ();

        setbuf (Reference_Buffer);
    }

    %   return to original position in reference buffer
    pop_spot ();

    %   anything found and listed?
    !if (ln)
    {
        occur_quit (o_ListBuffer);
        if (strlen(strtrim(str))) flush ("No match.");
        return;
    }

    %   initialise the occurrence list
    sw2buf (o_ListBuffer);
    bob ();
    go_down(3);
    fsearch(": ");
    occur_init (o_ListBuffer);
    set_buffer_modified_flag (0);
    set_readonly (1);
}


define occur(str)
%----------------------------------------------------------
%   This function does the real work finding and listing
%   occurrences of the argument.
%----------------------------------------------------------
{
    Reference_Buffer = whatbuf ();

    variable
        lnstr      = Null_String,
        ffound     = -1,
        ln_start   = -1,
        ln_ntmp    = -1,
        ln         = -1;

    ln_start = what_line();
    push_spot();
    bob();

    %   give list a name dependent on source
    o_ListBuffer =  "/O/ " + whatbuf ();

    %   get rid of any previous list
    sw2buf(o_ListBuffer);
    erase_buffer();

    str = substr (str, 1, 60);
    %   limit str to 60 charcaters

    occur_format (str);
    LAST_SEARCH = str;

    sw2buf (Reference_Buffer);

    bob();
    while (fsearch(str))
    {
        ln = what_column();
        bol();
        skip_white();

        push_mark ();
        eol ();

        %   to the stack
        bufsubstr ();

        ln = what_line();
        setbuf (o_ListBuffer);

        if (ln > 99999)
        {
            flush("Line number too big, exiting ...");
            break;
        }
        else vinsert ("%5d:  ", ln);

        %   insert from the stack
        insert (());
        newline ();

        setbuf (Reference_Buffer);
    }

    pop_spot();
    %   cursor in original position in reference buffer

    if(ln < 1)
    {
        flush("No Match.");
        delbuf (o_ListBuffer);
        return;
    }

    %   initialize list
    sw2buf (o_ListBuffer);
    bob ();

    %   position cursor in list
    %   according to cursor position in reference buffer
    do
    {
        bol();
        push_mark();
        call ("set_mark_cmd");
        go_right(5);
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        ln_ntmp = integer(lnstr);

        if (ln_ntmp < ln_start) continue;
        break;
    }
    while (down(1));

    occur_init (o_ListBuffer);
    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  %p  |  Reference Buffer: "+Reference_Buffer+"  |  F1=Help   ", 0);
    %%("  %p  |  Reference Buffer: "+Reference_Buffer+"  |  Keyword: "+ str+"  |  F1=Help  ", 0);

    %   use menu bar as title
    variable m = "&Occurrences:  " + str;
    dummy_menu (m);

    flush ("Press Enter to go to current entry in reference buffer.");
    %   useful in long lists where top is not seen
}


define occur_c_format_list ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   stub for C functions listing
%----------------------------------------------------------
{
    if (f_ListBuffer==Null_String) return;

    sw2buf (f_ListBuffer);

    erase_buffer ();
    newline ();
    loop (7) insert_char (' ');
    insert
    ("Press Enter on a line to go to same line in reference buffer.");
    newline ();
    newline ();
}


define occur_home ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves cursor to top of buffer
%----------------------------------------------------------
{
    bob();
}


define occur_end ()
%----------------------------------------------------------
%   ancillary function to occur ()
%   moves cursor to end of buffer, then back to last text
%----------------------------------------------------------
{
    eob();
    while(is_whiteline) occur_up();
}


define occur_comment_line ()
%---------------------------------------------------------
%   This function relies on the C tradition to start
%   intermediate comment lines with asterisk and space
%---------------------------------------------------------
{
    variable cn, found;

    bol ();
    skip_chars(" \t");
    cn = what_column ();

    found = ffind ("* ");
    if (found)
        if (cn == what_column ()) return 1;

    goto_column (cn);
    found = ffind ("/*");
    if (found)
        if (cn == what_column ()) return 1;

    goto_column (cn);
    found = ffind ("*/");
    if (found)
        if (cn == what_column ()) return 1;

    goto_column (cn);
    found = ffind ("//");
    if (found)
        if (cn == what_column ()) return 1;

    goto_column (cn);
    found = ffind ("%");
    if (found)
        if (cn == what_column ()) return 1;

    goto_column (cn);
    found = ffind ("#");
    if (found)
        if (cn == what_column ()) return 1;

    return 0;
}


define occur_sl_functions ()
%----------------------------------------------------------
%   Will list C or sl functions assuming the opening and closing
%   curly brackets are in column 1 - otherwise total failure.
%   Results are in general ok but sometime not exhaustive.
%   False positives will be avoided if comments do not contain
%   round brackets.
%----------------------------------------------------------
{
    %   no selection allowed (since it may get extended)
    if (markp)
    {
        flush("Not available when block highlighted.");
        return;
    }

    Reference_Buffer = whatbuf ();

    variable
        ln_near  = -1,
        ln_ntmp  = -1,
        ln       = -1,
        thisdir  = Null_String,
        thisfile = Null_String,
        thisfil1 = Null_String,
        lnstr    = Null_String,
        ccmnd    = Null_String;

    (thisfile, thisdir,,) = getbuf_info ();

    !if (strlen(thisfile))
    {
        flush("Save buffer first.");
        return;
    }

    ln_near = what_line ();
    push_spot ();
    %   we want the cursor to stay where it is

    thisfile = thisdir + thisfile;
    thisfil1 = thisfile + ".tmp1";

    %   give the function list a name dependent on source
    f_ListBuffer =  "/F/ " + whatbuf ();
    sw2buf(f_ListBuffer);
    erase_buffer();

    setbuf (Reference_Buffer);
    bob();

    CASE_SEARCH = 1;
    while (fsearch("define "))
    {
        eol();
        go_left(1);
        if (';' == what_char ()) continue;
        %   prototype skipped

        bol();
        %   skip the line beginning this search loop
        if(ffind("))")) continue;

        bol();
        %   skip comment lines containing search string
        %   37 is the comment charcater
        skip_white();
        if (37 == what_char ())
        {
            eol();
            continue;
        }

        %   entire line to the stack
        push_mark ();
        eol();
        bufsubstr();

        ln = what_line();
        setbuf (f_ListBuffer);
        if (ln > 99999)
        {
            flush("Line number too big, exiting ...");
            break;
        }
        else vinsert ("%5d:  ", ln);

        %   insert from the stack
        insert (());
        newline ();
        setbuf (Reference_Buffer);
    }
    CASE_SEARCH = 0;

    pop_spot();
    %   cursor back to original position in reference buffer

    setbuf (f_ListBuffer);
    bob();
    newline ();
    loop (8) insert_char (' ');
    insert
    ("Press Enter on a line to go to same line in reference buffer.");
    newline ();
    newline ();
    () = fsearch(":  ");

    %   position cursor in list
    %   according to cursor position in reference buffer
    do
    {
        bol();
        push_mark();
        call ("set_mark_cmd");
        go_right(5);
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        ln_ntmp = integer(lnstr);

        if (ln_ntmp < ln_near) continue;
        break;
    }
    while (down(1));

    occur_init (f_ListBuffer);
    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  %p  |  Functions in: "+Reference_Buffer+"  |  F1=Help  ", 0);

    variable m = "&Functions";
    dummy_menu(m);
}


define occur_functions ()
%----------------------------------------------------------
%   Will list functions using the utility ctags or routing
%   to another function for SLang (which is unknown to ctags).
%   It works for all languages covered by ctags but it works
%   using disk files. Therefore if the current buffer is
%   not saved there will be discrepancies.
%----------------------------------------------------------
{
    Reference_Buffer = whatbuf ();

    if (is_substr (Reference_Buffer, ".sl"))
    {
        occur_sl_functions();
        return;
    }

    variable
        ln_near  = -1,
        ln_ntmp  = -1,
        thisdir  = Null_String,
        thisfile = Null_String,
        thisfil1 = Null_String,
        lnstr    = Null_String,
        ccmnd    = Null_String;

    (thisfile, thisdir,,) = getbuf_info ();


    if (0==strlen(thisfile) or 0!=buffer_modified ())
    {
        flush("Save buffer first.");
        return;
    }

    !if (is_substr (Reference_Buffer, ".c") or is_substr (Reference_Buffer, ".h"))
    {
        flush("Is this a source code file (y/n)?");
        variable sstr=Null_String;
        sstr = read_mini2 ("Is this a source code file (y/n)?", "n", "n");
        if (sstr!="y" and sstr!="Y") return;
    }

    %%modf = buffer_modified ();
    ln_near = what_line ();

    thisfile = thisdir + thisfile;
    thisfil1 = thisfile + ".tmp1";

    %   give the function list a name dependent on source
    f_ListBuffer =  "/F/ " + whatbuf ();
    sw2buf(f_ListBuffer);

    %   let ctags do the chores in raw form
    ccmnd = "ctags -x " + thisfile + " > " + thisfil1;

    () = run_shell_cmd (ccmnd);

    %   ctags may issue warnings and dirty the screen
    !if (MINIBUFFER_ACTIVE) reset_screen ();

    erase_buffer();
    ()= insert_file (thisfil1);
    delete_file (thisfil1);
    bob ();

    %   clean up the list
    while (fsearch(thisfile))
    {
        % get rid of the source file name in list, not neded
        ed_delete_word();

        % get rid of abbreviated function name, not needed
        bol();
        ed_delete_word();
    }

    %   some formatting
    bob();
    do
    {
        %   get the line number into a string
        push_mark();
        call ("set_mark_cmd");
        () = ffind(" ");
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        %   delete the line number
        bol();
        ed_delete_word();

        %   insert formatted line number
        if (integer(lnstr)> 99999)
        {
            flush ("Line number too big, exiting ...");
            break;
        }
        else if (strlen(lnstr)) vinsert ("%5d:  ", integer(lnstr));
    }
    while (down(1));

    %   sort list by line number
    bob();

    if (eobp==bobp)
    {
        delbuf(whatbuf);
        flush("Not found.");
        return;
    }

    toggle_block (1);
    eob();
    while (is_whiteline) go_up(1);
    bol();
    go_right(4);
    update(0);
    sort1(0);
    toggle_block (0);
    bob();

    newline ();
    loop (7) insert_char (' ');
    insert
    ("Press Enter on a line to go to same line in reference buffer.");
    newline ();
    newline ();
    () = fsearch(":  ");

    %   position cursor in list
    %   according to cursor position in reference buffer
    do
    {
        bol();
        push_mark();
        call ("set_mark_cmd");
        go_right(5);
        lnstr = strtrim(bufsubstr());
        %   mark is popped

        ln_ntmp = integer(lnstr);

        if (ln_ntmp < ln_near) continue;
        break;
    }
    while (down(1));

    occur_init (f_ListBuffer);
    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  %p  |  Functions in: "+Reference_Buffer+"  |  F1=Help  ", 0);

    variable m = "&Functions";
    dummy_menu(m);
    flush (" ");
}


define occur_shiftkeys_show ()
%----------------------------------------------------------
%   ancillary to occur_shiftkeys ()
%----------------------------------------------------------
{
    %   only proceed on valid lines
    bol ();
    !if (ffind(":  ")) return;

    %   buffers involved
    b_tmpbf = whatbuf ();
    variable tmp_bf = " /shiftkey_details/";

    %   get the line number from begin of line
    bol ();
    push_mark ();
    !if (ffind(":  ")) return;
    push_spot();
    variable lline = integer (bufsubstr ());
    %   mark popped here

    %   get the file name

    goto_column(9);
    push_mark();
    skip_word_chars();
    variable ffile = bufsubstr();
    ffile = strcat (JED_ROOT+"/lib/",ffile);

    %   load file, find line, show it
    sw2buf (tmp_bf);
    erase_buffer();

    insert_file (ffile);
    goto_line (lline);
    go_up(3);
    newline();
    newline();
    insert("line ");
    insert(string(lline));
    insert(",  ");
    insert(ffile);
    newline();
    () = bsearch("line ");
    recenter (2);
    () = fsearch("Shift not pressed:");
    hide_status_indicators(1);
    set_status_line("  Press any key to exit  ",0);
    variable m = "&Shiftkeys Details";
    dummy_menu(m);

    update_sans_update_hook (0);

    %   wait for a key to close
    flush_input(getkey ());

    sw2buf(b_tmpbf);
}


define occur_shiftkeys ()
%----------------------------------------------------------
%   special list: shift keys in use but not in a keymap
%----------------------------------------------------------
{
    %   open temporary buffer
    variable b_tmpbf = "/Shiftkey Commands/";
    variable cmdstr = Null_String;
    variable fmask = strcat (JED_ROOT,"/lib/*.sl");
    sw2buf(b_tmpbf);
    erase_buffer ();

    cmdstr = "grep -ins 'Shift not pressed:' " + fmask + " 2> /dev/nul &";

    () = run_shell_cmd (cmdstr);
    %   output goes into current buffer

    bob();
    while (fsearch("Shift not pressed:"))
        replace_chars (strlen("Shift not pressed:"),"");
    bob();
    while (fsearch(":")) replace_chars (1," ");
    bob();

    variable i=0, j=0, lastln=16, ln=0;
    variable bfilez = String_Type [lastln];
    variable ffilez = String_Type [lastln];
    variable llinez = String_Type [lastln];
    variable hhintz = String_Type [lastln];

    bfilez [*] = Null_String;
    ffilez [*] = Null_String;
    llinez [*] = Null_String;
    hhintz [*] = Null_String;

    %   load hits into array for further operations
    while (fsearch(".sl"))
    {
        if (i > lastln-1) break;

        bol();
        !if(ffind("%   "))
        {
            del_to_eol();
            continue;
            %   irrelevant line
        }
        bol();

        %   file name

        push_mark ();
        skip_word_chars ();
        ffilez[i] = bufsubstr ();
        bfilez[i] = path_basename(ffilez[i]);

        %   line number

        skip_non_word_chars ();
        push_mark();
        skip_word_chars ();
        llinez[i] = bufsubstr();

        %   explanatory text

        () = ffind("%");
        go_right(1);
        skip_non_word_chars ();
        push_mark();
        eol();
        hhintz[i] = bufsubstr();

        i++;
    }
    if (i < lastln) lastln=i;

    if (i==0)
    {
        flush ("List is empty");
        delbuf(whatbuf);
        return;
    }

    erase_buffer ();

    %   header text for list
    newline ();
    goto_column(9);
    variable instruct = "Press Enter on a line to see details.";
    insert (instruct);
    newline ();
    newline ();

    for (i=0; i<lastln; i++)
    {
        ln = integer(llinez[i]);

        if (ln > 99999)
        {
            flush("Line number too big, skipping ...");
            sleep(1);
            continue;
        }
        else vinsert ("%5s:  ",llinez[i]);

        goto_column(9);
        insert (bfilez[i]);
        goto_column(34);
        insert (hhintz[i]);
        eol();
        newline();
    }
    bsearch (instruct);

    %   prepare for the list
    set_readonly(1);
    set_buffer_modified_flag (0);
    set_status_line ("  Escape to exit  ", 0);
    variable m = "&Shiftkey Commands (not in keymaps)";
    dummy_menu(m);

    occur_init ();

    %   will use own keymap
    reset_slim_keymap ();
    $1 = "shiftkeys_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    definekey ("occur_shiftkeys_show",  Key_Enter,  $1);

    recenter(2);
}


define occur_list (rx)
%----------------------------------------------------------
%   0   list occurrences case-insensitive
%   1   list occurrences of regular expression
%----------------------------------------------------------
{
    variable str        = Null_String;
    variable tmpbf      = whatbuf ();

    %   use block on a single line as key, if any

    if (markp ())
    {
        check_stream ();
        if (what_block ()==0 and BegRow==EndRow)
            str = bufsubstr ();
        %   block on one line only

        toggle_block (0);
    }

    variable o_prompt;
    if (rx==1)
    {
        o_prompt = "Regex for occurrences: ";
    }

    else
    {
        o_prompt = "Case-insensitive occurrences of:";
        rx=0;
    }

    if (str == Null_String)
        str = read_mini2 (o_prompt, Null_String, Null_String);

    if (LASTKEY[0] == 23) return;
    %   discard reply ^w, ascii 23

    !if (strlen(str)) return;

    if (rx) occur_rx(str);
    else occur(str);
}



%**********************************************************
%   Functions to list buffers and switch buffer
%
%   dependencies:
%   bufed
%     bufed_list
%       bufed_buffers
%     bufed_create_keymap
%       bufed_sw2buf
%         bufed_get
%
%**********************************************************


define bufed_buffers ()
%----------------------------------------------------------
%   Will list buffer currently opened and accessible to user
%----------------------------------------------------------
{
    variable i, j, tmp, this, name, flags, flag_chars;
    variable umask;
    variable name_col, dir_col, mode_col;

    mode_col = 14;
    name_col = 21;
    dir_col  = 45;

    tmp = "/BufferList/";
    this = whatbuf ();
    sw2buf(tmp);
    set_readonly(0);
    erase_buffer ();

    set_status_line
    ("  Press Enter on highlighted buffer to jump to buffer  |  F1=Help  ", 0);

    variable m = "&Buffer List";
    dummy_menu(m);

    newline ();

    flag_chars = "CBKN-UORDAM";
    insert ("  Flags");
    goto_column (mode_col);
    insert ("umask");
    goto_column (name_col);
    insert ("Buffer");
    goto_column(dir_col);
    insert("Dir/File");
    newline ();
    newline ();

    loop (buffer_list ())
    {
        name = ();
        if (int(name) == ' ') continue;
        if (int(name) == '/') continue;
        %   internal non-editable buffers not to be listed

        flags = getbuf_info (name);

        % more on stack

        umask = set_buffer_umask (-1);
        bol ();
        i = 0x400;
        j = 0;
        while (i)
        {
            if (flags & i) flag_chars[j];
            else '-';
            insert_char (());
            i = i shr 1;
            j++;
        }
        goto_column (mode_col);
        vinsert ("0%03o", umask);
        goto_column (name_col);

        % Buffername may contain whitespace, enclose it in quotes
        insert_char ('"');
        insert(());
        %   buffer name inserted

        insert_char ('"');

        goto_column(dir_col-2);
        !if (eolp ())
        {
            eol ();
            insert(" ");
        }

        insert(());
        insert(());
        %   dir/file inserted

        newline ();
    }
}


define bufed_list ()
%----------------------------------------------------------
%   Prepares a template for the buffers list
%----------------------------------------------------------
{
    variable Bufed_buf = "/BufferList/";

    check_buffers ();
    bufed_buffers ();
    sw2buf (Bufed_buf);

    bob ();
    newline ();

    variable line1 =
    "A: Autosave             B: Binary File      C: CRmode  ";

    variable line2 =
    "D: Disk File Changed    K: Not backed up    M: Modified";

    variable line3 =
    "N: No autosave          O: Overwrite        R: Readonly    U: Undo";

    variable underln =
    "__________________________________________________________________";

    variable rmargin = "      ";

    newline ();
    insert (rmargin);
    insert (line1);
    newline ();
    insert (rmargin);
    insert (line2);
    newline ();
    insert (rmargin);
    insert (line3);
    newline();
    insert (rmargin);
    insert (underln);
    newline ();
    newline ();

    go_down (1);
}


define bufed_get ()
%----------------------------------------------------------
%   Will get the buffer name associated with the current line
%   in buffers list
%----------------------------------------------------------
{
    EXIT_BLOCK  pop_spot ();

    variable buf;

    push_spot ();
    bol ();

    !if (ffind_char ('"')) return Null_String;

    go_right (1);
    push_mark ();
    !if (ffind_char ('"'))
    {
        pop_mark (1);
        return Null_String;
    }

    buf = bufsubstr ();
    !if (bufferp (buf)) buf = "";
    return buf;
}


define bufed_sw2buf ()
%----------------------------------------------------------
%   Will switch to the buffer associated with the current line
%   in buffers list
%----------------------------------------------------------
{
    variable buf = bufed_get ();
    !if (int (buf)) return;
    if (buf == "/BufferList/") return;

    sw2buf ("/BufferList/");
    set_buffer_modified_flag (0);
    delbuf (whatbuf);

    sw2buf (buf);
}


define bufed_down ()
%----------------------------------------------------------
%   Down one line in the buffers list
%----------------------------------------------------------
{
    !if(down(1)) return;
    bol ();
    !if (ffind("\"")) go_up(1);
}


define bufed_up ()
%----------------------------------------------------------
%   Up one line in the buffers list
%----------------------------------------------------------
{
    !if(up(1)) return;
    bol ();
    !if (ffind("\"")) go_down(1);
}


define bufed_help ()
%----------------------------------------------------------
%   Provides help for the buffers list
%----------------------------------------------------------
{
    variable
        stringA = "BufferList",
        stringB = "    [#]",
        stringC = "&Quick Help: Buffer List";

    display_hlp2 (stringA,stringB,stringC);
}


define bufed_create_keymap ()
%----------------------------------------------------------
%   Creates a keymap for the buffers list
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "viewer_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<127; i++)
        undefinekey (char(i), $1);

    definekey ("bufed_sw2buf ()",   Key_Enter,  $1);
    definekey ("bufed_down",        Key_Down,   $1);
    definekey ("bufed_down",        Key_Right,  $1);
    definekey ("bufed_up",          Key_Up,     $1);
    definekey ("bufed_up",          Key_Left,   $1);
    definekey ("bufed_help",        Key_F1,     $1);
}


define bufed_start ()
%----------------------------------------------------------
%   Initialize the buffers list
%----------------------------------------------------------
{
    variable this_buf = sprintf ("\"%s\"", whatbuf ());
    bufed_list ();
    () = fsearch (this_buf);

    bufed_create_keymap ();

    set_readonly (1);
}


%**********************************************************
%   Functions for MINIBUFFER operations
%   incl. recall of previous commands
%   trim and only save commands of at least 4 characters
%   save only commands different from old ones
%   two separate history buffers are used
%       0   for sundry strings
%       1   for paths and files
%   switching between the two buffers is done in ledit.c
%   read_with_completion calls history buffer 1
%   everything else calls history buffer 0
%   the intrinsic read-only variable that points to the
%   history buffers is mS, called Mini_Store in the C source
%**********************************************************

%   Mini_Top is the max. number of lines allowed for history
%   Mini_History is a string array for history lines
%       separated for sundry and files
%   Mini_Last is the index of line reached when recalling history

variable
    Mini_Top = 16,
    Mini_History = String_Type [2, Mini_Top],
    Mini_Last = Integer_Type [2];

Mini_Last[*] = 0;
Mini_History[*,*] = Null_String;

define mini_test ()
{
}

define mini_next ()
%----------------------------------------------------------
%   Get next command from minibuffer's history
%   'next' means decreased index
%   cycle around at bottom
%----------------------------------------------------------
{
    erase_buffer ();

    variable ndx = Mini_Last[mS]-1;
    if (ndx < 0) ndx += Mini_Top;

    while (strlen(Mini_History[mS, ndx])==0 and ndx > 0)
        ndx--;

    if (strlen(Mini_History[mS, ndx]))
    {
        insert (Mini_History[mS,ndx]);
        Mini_Last[mS] = ndx;
    }
    else
    {
        Mini_Last[mS]=0;
        return;
    }
}


define mini_prev ()
%----------------------------------------------------------
%   Get previous command from minibuffer's history
%   'previous' means increased index
%   cycle around at top
%----------------------------------------------------------
{
    erase_buffer ();

    if (strlen(Mini_History[mS, Mini_Last[mS]]))
        insert (Mini_History[mS, Mini_Last[mS]]);
    else
    {
        Mini_Last[mS]=0;
        return;
    }

    Mini_Last[mS]++;
    Mini_Last[mS] = Mini_Last[mS] mod Mini_Top;
}


define mini_storeline (s)
%----------------------------------------------------------
%   Will store current (accepted) line in minibuffer's
%   history.
%----------------------------------------------------------
{
    %   s is already strtrim(s)
    %   there is no whitespace at both ends, already trimmed

    %   keep short lines out of history
    if (strlen(s) < 4) return;

    %   insert argument s at begin of history,
    %   increase index for older entries discarding top
    %   a temp array is used for operations

    variable i=0, j=0, strchk=0;
    variable Mini_Tmp = String_Type [2, Mini_Top];
    Mini_Tmp [*,*] = Null_String;

    Mini_Tmp [mS,j] = s;
    j++;
    %   slang error if index incremented directly in array!!!!!

    %   we copy to temp array history entries not empty
    %   and not overlapping with argument s
    %   start with index 1 because index 0 is taking fresh entry

    for (i=0; i<Mini_Top; i++)
    {
        if (j==Mini_Top) break;
        if(Mini_History[mS,i] == Null_String)
            continue;

        strchk = is_substr(s, Mini_History[mS,i]);
        %   neither argument is Null_String

        %   if entry has no overlap with argument s
        %   keep it
        !if (strchk)
        {
            Mini_Tmp [mS, j] = Mini_History[mS,i];
            j++;
        }
    }

    %   array Mini_Tmp has now the refreshed history
    %   no repetitions, no empty strings

    Mini_History[mS, *] = Mini_Tmp[mS, *];
}


define mini_remove ()
%----------------------------------------------------------
%   Will remove from history the currently displayed line
%----------------------------------------------------------
{
    variable i=0, c_line = Null_String;
    c_line = strtrim(line_as_string);
    erase_buffer ();

    !if(strlen(c_line)) return;

    mini_storeline (c_line);

    %   that line is now at begin of history array
    %   we discard the first element of array

    for (i=0; i < Mini_Top-1; i++)
        Mini_History[mS,i] = Mini_History[mS,i+1];

    Mini_Last[mS] = 0;
}


define mini_quit ()
%----------------------------------------------------------
%   Will quit minibuffer passing on nothing
%----------------------------------------------------------
{
    call ("exit_mini");
    Mini_Last[mS] = 0;
}


define mini_exit ()
%----------------------------------------------------------
%   Will exit the minibuffer passing on the current line.
%----------------------------------------------------------
{
    variable c_line = Null_String;
    c_line = strtrim(line_as_string);
    mini_storeline (c_line);
    call ("exit_mini");
    Mini_Last[mS] = 0;
}


define mini_paste ()
%----------------------------------------------------------
%   Will paste from the clipboard in the minibuffer
%----------------------------------------------------------
{
    call("yank");
}


define mini_copy ()
%----------------------------------------------------------
%   Will copy entire minibuffer to clipboard
%----------------------------------------------------------
{
    push_spot ();
    bob();
    push_mark ();
    eob ();
    if (eobp==bobp) return;
    call ("copy_region");
    % mark gone
    pop_spot ();
}


define mini_del_bol ()
%----------------------------------------------------------
%   Will delete to begin of line in the minibuffer
%----------------------------------------------------------
{
    push_mark ();
    bol ();
    del_region();
}


define mini_del_eol ()
%----------------------------------------------------------
%   Will delete to end of line in the minibuffer
%----------------------------------------------------------
{
    push_mark ();
    eol ();
    del_region();
}


define mini_functionhelp ()
{
    %   not used yet, shape of things to come

    variable
        stringA = _function_name (),
        stringB = "F1",
        stringC = "&MiniBuffer Current Function";

    variable tmpbf = whatbuf ();

    display_hlp2 (stringA,stringB,stringC);

    pop2buf(tmpbf);
    %   whithout this command, editor will crash
}


define mini_quickhelp ()
{
    variable
        stringA = "MiniBuffer",
        stringB = " ",
        stringC = "&MiniBuffer Keybindings",
        tmpbf   = whatbuf ();

    display_help(stringA,stringB,stringC);
    %   display_hlp2 bad idea

    otherwindow();
    %   without this command, editor will crash
}


define mini_init_minibuffer ()
%----------------------------------------------------------
%   Will set up the minibuffer,
%   called only once from site.sl
%----------------------------------------------------------
{
    variable i;
    variable mini = "Mini_slim_Map";
    %   must be same name as in ledit.c

    !if (keymap_p (mini)) make_slim_keymap (mini);

    for (i=0; i<27; i++) undefinekey (char(i), mini);

    %definekey ("mini_test",             "^T",           mini);
    definekey ("mini_copy",             "^C",           mini);
    definekey ("mini_paste",            "^V",           mini);
    definekey ("mini_quit",             "^W",           mini);
    definekey ("mini_remove",           "^Y",           mini);
    definekey ("mini_del_bol",          Key_Ctrl_K_BS,  mini);
    definekey ("mini_del_eol",          Key_Ctrl_K_Del, mini);
    definekey ("quoted_insert",         Key_Ctrl_K_Ins, mini);
    definekey ("mini_next",             Key_Down,       mini);
    definekey ("eob",                   Key_End,        mini);
    definekey ("mini_exit",             Key_Enter,      mini);
    definekey ("mini_quickhelp",        Key_F1,         mini);
    definekey ("bob",                   Key_Home,       mini);
    definekey ("toggle_overwrite",      Key_Ins,        mini);
    definekey ("mini_prev",             Key_Up,         mini);

    %   additionally available
    %   backward_delete_char    backspace       ledit.c
    %   delete_char_cmd         delete          ledit.c
    %   kbd_quit                ctrl-backslash  keymap.c
}


%**********************************************************
%
%   Function to run gcc and check compile errors
%
%   gcc will run in a subshell.
%
%   The compile function will issue a command to gcc
%   from inside the editor. An output buffer for the compiler
%   will be created with line numbers referencing errors
%   in the source files. Editing the source files, may well
%   invalidate the line references.
%
%   The variable Compile_Output_Buffer is defined in site.sl
%
%**********************************************************



define compile ()
%----------------------------------------------------------
%   Runs the compiler directing messages to an editor buffer
%----------------------------------------------------------
{
    variable
        cmd=Null_String,
        dir=Null_String,
        status,
        tmp=Null_String;

    %   initialize compiler output buffer
    sw2buf(Compile_Output_Buffer);
    set_readonly(0);
    erase_buffer ();

    %   get current directory
    (,dir,,) = getbuf_info ();

    %   current directory can be changed
    %   before issuing compile command
    forever
    {
        eob();
        newline();
        insert ("Current Directory:    ");
        insert(dir);
        %newline();

        tmp = read_path_from_mini ("Compile from directory:");
        %   will go to sundry history buffer

        if (LASTKEY[0] == 23) tmp=Null_String;
        %   discard reply ^w, ascii 23

        !if(strlen(tmp))
        {
            delbuf(whatbuf());
            return;
        }

        if (chdir(tmp))
        {
            flush ("Directory not found, press any key...");
            flush_input(getkey ());
            flush(" ");
        }
        else break;
    }

    %   display the current directory for the compile command
    eob();
    newline();
    insert ("Current Directory:    ");
    dir = tmp;
    insert(dir);
    newline();
    newline();

    %   enter the compile command
    cmd = read_mini2 ("Compile Command:","", "make");

    !if (strlen (cmd))
    {
        delbuf(whatbuf());
        return;
    }

    %   still in compiler output buffer
    %   send the compile command to the shell

    variable linA = "Compiling from:       " + dir;
    variable linB = "Compile command:      " + cmd;
    variable lin1 = "Control will revert to user when the compiler is done.";
    variable lin2 = "However, editor can also be started in another console.";
    insert(linA);
    newline();
    insert(linB);
    newline();
    newline();
    insert(lin1);
    newline();
    insert(lin2);
    newline();
    newline();
    insert ("Please wait...");
    newline();
    newline();

    update_sans_update_hook (0);

    cmd = sprintf ("(%s) 2>&1 < /dev/null", cmd);
    status = run_shell_cmd (cmd);

    %   add exit status and warning to the buffer
    newline ();
    cmd = sprintf ("Exit Status: %d", status);
    insert(cmd);

    if (status != 0)
    {
        eob();
        newline();
        newline();
        tmp = "Editing source may invalidate line numbers in this list.";
        insert(tmp);
    }
}


%**********************************************************
%   Recent Files
%   A list of recently edited files is kept in
%   the user's home directory.
%   The list can be browsed to open a file again.
%   Files are appended to it when they are saved.
%**********************************************************


define recent_quit ()
%----------------------------------------------------------
%   Close the recent files list with a key command
%   directly from the list itself.
%   Don't call from other functions.
%----------------------------------------------------------
{
    set_buffer_modified_flag (0);
    set_readonly (0);
    delbuf (whatbuf);
    if (bufferp(b_tmpbf)) sw2buf(b_tmpbf);
    else prevbuffer();
}


define recent_remove_line ()
%----------------------------------------------------------
%   will remove a line from the list of recently edited files
%   call from keyboard, don't call from other functions
%----------------------------------------------------------
{
    if (globl_str=="&Recently Accessed Directories")
        return;

    bol();
    !if (ffind("/")) return;

    set_readonly (0);
    bol();
    del_to_eol ();
    call ("delete_char_cmd");
    bob();

    %   any entry?
    if (fsearch("/"))
    {
        %   save region
        push_mark();
        eob();
        write_region_to_file (RecentFiles);
        bob();
        () = fsearch("/");
        set_readonly (1);
    }

    else
    {
        () = delete_file(RecentFiles);
        delbuf(whatbuf);
        if (bufferp(b_tmpbf)) sw2buf(b_tmpbf);
    }
}


define recent_open_file ()
%----------------------------------------------------------
%   Shift not pressed:  will open a file from the list
%                       of recently edited files
%
%   Shift pressed:      will open only path from the list
%                       of recently edited files
%----------------------------------------------------------
{
    if (4 > what_line()) return;

    variable recentln = Null_String;

    bol ();
    push_mark();
    eol();
    recentln = strtrim_beg(bufsubstr());

    !if (strlen(recentln)) return;
    if (file_status (recentln) != 1)
    {
        flush ("Not found.");
        return;
    }

    delbuf(whatbuf);

    %   press shift to open only the path of highlighted file
    if (SHIFTKEY_PRESSED)
    {
        variable recentdir = Null_String;
        recentdir  = path_dirname(recentln);
        dls (recentdir);
    }
    else
    {
        () = find_file(recentln);
        b_tmpbf = Null_String;
    }
}


define recent_open_dir ()
%----------------------------------------------------------
%   will open a directory from the list of recently accessed
%----------------------------------------------------------
{
    if (4 > what_line()) return;

    variable recentln = Null_String;

    bol ();
    push_mark();
    eol();
    recentln = strtrim_beg(bufsubstr());

    !if (strlen(recentln)) return;
    if (file_status (recentln) != 2)
    {
        flush ("Not found.");
        return;
    }

    delbuf(whatbuf);

    variable dummy_file = recentln + " ";
    () = read_file (dummy_file);
    %   will bring the directory to top of recent list
    %   even if no file opened there
    delbuf(whatbuf);

    dls (recentln);
}


define recent_load ()
%----------------------------------------------------------
%   open file or directory from the list of recently used
%----------------------------------------------------------
{
    variable m = globl_str;
    globl_str = Null_String;

    if (m == "&Recently Accessed Directories")
    {
        recent_open_dir ();
        return;
    }

    recent_open_file ();
}


define recent_init ()
%----------------------------------------------------------
%   defines the keymap for the list of recent files/dirs
%----------------------------------------------------------
{
    reset_slim_keymap ();
    $1 = "recent_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    variable i;
    for (i=32; i<255; i++) undefinekey (char(i), $1);
    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
    undefinekey (Key_Home,  $1);
    undefinekey (Key_End,   $1);
    undefinekey (Key_Del,   $1);
    undefinekey (Key_BS,    $1);
    undefinekey ("^S",      $1);

    definekey ("recent_remove_line",    Key_Del,    $1);
    definekey ("recent_load",           Key_Enter,  $1);
    definekey ("recent_quit",           "^W",       $1);
}


define recent_files_update ()
%----------------------------------------------------------
%   Will pre-pend the current file name to the list of
%   recently edited files (disk file).
%   Update is dependent on saving the current file,
%   files opened but not saved will not be considered.
%----------------------------------------------------------
{
    variable fst = file_status (RecentFiles);
    if (fst!=0 and fst!=1) return;

    variable RecentAdd = Null_String;
    variable basebf    = Null_String;

    RecentAdd = buffer_filename ();
    %   file associated with the buffer just saved

    variable llen = strlen(RecentAdd);

    %   better safe than sorry
    !if(llen) return;

    if (RecentAdd == RecentFiles) return;
    if (RecentAdd == RecentDirs) return;
    %   disregard the two lists of recent files and dirs
    %   they cannot be added to themselves
    %   they would segfault because of the hook

    %   get rid of backups terminating with tilde
    if ("~" == substr(RecentAdd,llen,1)) return;

    %   get rid of tmp files
    if (is_substr(RecentAdd, ".tmp")) return;

    %   red is an associative array to indicate
    %   whether a file has already been added
    %   to the recent list in this editor session.
    %   if key exists, the same file can be
    %   saved and reloaded in the same session
    %   but only one entry will go for it to the recent list

    if (assoc_key_exists (red, whatbuf)) return;
    else red[whatbuf]=1;

    %   acrobatics to pre-pend entry to RecentFiles
    %   don't use find_file or read_file
    %   because hooks are mingling

    sw2buf (RecentzBuf);
    () = insert_file (RecentFiles);
    bob();
    insert(RecentAdd);
    newline();
    () = write_buffer (RecentFiles);
    flush (" ");
    %   buffer is renamed by saving

    %   mistery: without following workaround
    %   cursor may jump in saved file
    basebf = path_basename(RecentAdd);
    if (bufferp(basebf)) sw2buf(basebf);

    basebf = path_basename(RecentFiles);
    if (bufferp(basebf)) delbuf (basebf);

    %   The name of file just saved
    %   is inserted at top of recent list
}
add_to_hook ("_jed_save_buffer_after_hooks", &recent_files_update);


define recent_dirs_update ()
%----------------------------------------------------------
%   Will add the current directory to the list of recently
%   accessed dirs. Update is dependent on opening an existing
%   disk file and so accessing a directory. Access of same
%   directory is recorded only once per session.
%----------------------------------------------------------
{
    variable fst = file_status (RecentDirs);
    if (fst < 0) return;
    if (fst > 1) return;

    variable
        ddir = Null_String,
        bf   = Null_String,
        bn   = Null_String;

    %   in current buffer is the file just opened with find_file

    (bf, ddir, bn,) = getbuf_info ();

    if (bf == Null_String)          return;
    if (is_substr(bn, "/"))         return;
    if (is_substr(RecentFiles,bn))  return;
    if (is_substr(RecentDirs,bn))   return;

    if (file_status (ddir) != 2)
    {
        delbuf (bf);
        flush ("Invalid directory: " + ddir);
        return;
    }
    %   it can happen when find_file creates new buffer
    %   for a file not yet on disk

    %   r_dir is an associative array to indicate
    %   whether a dir has already been added
    %   to the recent list in this editor session.
    %   if key exists, no further entry for the dir
    %   will go to the recent list

    if (assoc_key_exists (r_dir, ddir)) return;
    else r_dir[ddir]=1;

     %   acrobatics to pre-pend new entry to RecentDirs

    sw2buf (RecentzBuf);
    erase_buffer();
    () = insert_file (RecentDirs);
    bob();
    insert(ddir);
    newline();
    () = write_buffer (RecentDirs);
    %   buffer is renamed by saving
    flush (" ");
    variable basebf = path_basename(RecentDirs);
    if (bufferp(basebf)) delbuf (basebf);

    %   The directory of the file just opened
    %   is inserted at top of recent list
    %
    if (bufferp(bn)) sw2buf(bn);
}
add_to_hook ("_jed_find_file_after_hooks", &recent_dirs_update);


define recent_files ()
%----------------------------------------------------------
%   This will display a list of max. 16 recently edited files
%   and allow opening them.
%----------------------------------------------------------
{
    if (1 !=  file_status (RecentFiles))
    {
        flush("No list of recently edited files.");
        return;
    }

    %   if the recent files list is being edited, close it
    variable basebf = path_basename(RecentFiles);
    if (bufferp(basebf)) delbuf (basebf);

    b_tmpbf = whatbuf ();
    %   buffer from which the recent list is called
    %   variable defined in this module across functions

    sw2buf (RecentzBuf);
    erase_buffer();
    () = insert_file (RecentFiles);
    %   this has the advantage of no queries when saving
    bob();
    newline ();
    bob();

    !if (fsearch ("/"))
    {
        delbuf(whatbuf);
        flush("No list of recently edited files.");
        sw2buf(b_tmpbf);
        return;
    }
    bob();

    variable i=0, j=0, lastln=32;
    variable ffilez = String_Type [lastln];
    ffilez [*] = Null_String;

    %   load recent files into array for further operations
    while (fsearch("/"))
    {
        if (i > lastln-1) break;
        push_mark ();
        eol();
        ffilez[i] = bufsubstr ();
        i++;
    }
    if (i < lastln) lastln=i;

    %   mark repeated entries (turn into Null_String)
    %   same for files previously edited but not avalable any longer
    for (i=0; i<lastln; i++)
    {
        if (ffilez[i]==Null_String) continue;

        if (file_status (ffilez[i]) != 1)
        {
            ffilez[i] = Null_String;
            continue;
        }

        for (j=i+1; j<lastln; j++)
        {
            if (ffilez[i] == ffilez[j]) ffilez[j] = Null_String;
        }
    }

    %   shrink the list, drop empty lines
    ffilez = ffilez[where(ffilez!=Null_String)];

    variable llen = length(ffilez);

    if (llen==0)
    {
        delbuf(whatbuf);
        flush("No list of recently edited files.");
        sw2buf(b_tmpbf);
        return;
    }

    if (llen > 16) llen = 16;

    %   some administration to keep list tidy on disk

    erase_buffer ();
    for (i=0; i < llen ; i++)
    {
        insert (ffilez[i]);
        newline();
    }
    () = write_buffer (RecentFiles);
    %   buffer is renamed by saving
    flush (" ");
    if (bufferp(basebf)) delbuf (basebf);

    %   now create pick list

    sw2buf (RecentzBuf);
    erase_buffer();
    newline();
    goto_column (5);
    insert ("Press Enter on highlighted file to open it.");
    newline ();
    goto_column (5);
    insert ("[ Shift-Enter to open path only. ]");
    newline ();
    newline ();

    %   insert file names

    for (i=0; i < llen; i++)
    {
        insert ("    ");
        insert (ffilez[i]);
        newline();
    }

    %   recent list is in special buffer

    bob();
    () = fsearch("Press");
    recent_init();

    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  Escape to exit  |  Up/Down to navigate  |  Del to remove entry  ", 0);

    %   hide the editing menu bar
    variable m = "&Recently Edited Files";
    dummy_menu(m);

    %   set globl_str to tell files list from directories list
    globl_str = m;
    flush(" ");
}


define recent_dirs ()
%----------------------------------------------------------
%   Will display a list of max. 16 recently accessed directories
%   and allow opening them.
%----------------------------------------------------------
{
    if (1 !=  file_status (RecentDirs))
    {
        flush("No list of recently accessed directories.");
        return;
    }

    %   if the recent dirs list is being edited, close it
    variable basebf = path_basename(RecentDirs);
    if (bufferp(basebf)) delbuf (basebf);

    b_tmpbf = whatbuf ();

    (InsBuffer,,,) = getbuf_info ();
    %   this is needed should the recent directory
    %   be used to insert a file

    sw2buf (RecentzBuf);
    erase_buffer();
    () = insert_file (RecentDirs);
    bob();
    newline ();
    bob();

    !if (fsearch ("/"))
    {
        delbuf(whatbuf);
        flush("No list of recently accessed directories.");
        sw2buf(b_tmpbf);
        return;
    }
    bob();

    variable i=0, j=0, lastln=64;
    variable ffilez = String_Type [lastln];
    ffilez [*] = Null_String;

    %   load entries into array for further operations
    while (fsearch("/"))
    {
        if (i > lastln-1) break;
        push_mark ();
        eol();
        ffilez[i] = bufsubstr ();
        i++;
    }
    if (i < lastln) lastln=i;


    %   mark repeated entries (turn into Null_String)
    %   same for entries not avalable any longer
    for (i=0; i<lastln; i++)
    {
        if (ffilez[i]==Null_String) continue;

        if (file_status (ffilez[i]) != 2)
        {
            ffilez[i] = Null_String;
            continue;
        }

        for (j=i+1; j<lastln; j++)
        {
            if (ffilez[i] == ffilez[j]) ffilez[j] = Null_String;
        }
    }

    %   shrink the list, drop empty lines
    ffilez = ffilez[where(ffilez!=Null_String)];

    variable llen = length(ffilez);


    if (llen==0)
    {
        delbuf(whatbuf);
        flush("No list of recently accessed directories.");
        sw2buf (b_tmpbf);
        return;
    }

    if (llen > 16) llen = 16;

    %   some administration to keep list tidy on disk


    erase_buffer ();
    for (i=0; i < llen ; i++)
    {
        insert (ffilez[i]);
        newline();
    }
    () = write_buffer (RecentDirs);
    %   buffer is renamed by saving
    flush (" ");
    if (bufferp(basebf)) delbuf (basebf);


    %   now create pick list

    sw2buf (RecentzBuf);
    erase_buffer();
    newline();
    goto_column (5);
    insert ("Press Enter on highlighted directory to open it.");
    newline ();
    newline ();


    %   insert names

    for (i=0; i < llen; i++)
    {
        insert ("    ");
        insert (ffilez[i]);
        newline();
    }

    %   recent dirs in special buffer

    bob();
    () = fsearch("Press");
    recent_init();

    set_buffer_modified_flag (0);
    set_readonly (1);

    set_status_line
    ("  Escape to exit  |  Up/Down to move  ", 0);

    %   hide the editing menu bar
    variable m = "&Recently Accessed Directories";
    dummy_menu(m);

    %   set globl_str to tell apart files and directories list
    globl_str = m;
    flush(" ");
}


define c_mode()
{
    require("cmode.sl");
    set_mode("C", 2);
    WANT_SYNTAX_HIGHLIGHT = 1;
    use_syntax_table ("C");
}
