%**********************************************************
%   Directory listing for opening or inserting files.
%
%   Totally dependent on line format and sorting.
%   Any change to format implies revising the code.
%
%   Upon invocation, reply to the minibuffer prompt
%   is read and validated. If the entry identifies
%   one unique file for editing, this file is opened.
%   Otherwise a directory listing is displayed.
%   The list is scrollable and has its own menus.
%   Enter loads a file or opens a directory,
%   Backspace takes to upper directory.
%**********************************************************


%**********************************************************
%   Prototypes
%**********************************************************

define dls(dir);
define dls_align_right(N,L);
define dls_copy_all();
define dls_copy_line();
define dls_create_menu();
define dls_delete_file();
define dls_display_help();
define dls_down();
define dls_draw_list(whichdir);
define dls_end();
define dls_format_time(sstt);
define dls_getfile();
define dls_grep();
define dls_help();
define dls_home();
define dls_init_keymap();
define dls_insert_current();
define dls_insert_from_listing();
define dls_jump_to_initial();
define dls_list_keys();
define dls_load(name);
define dls_lynx();
define dls_man();
define dls_menubar();
define dls_octal_to_letters(o_string);
define dls_open_current();
define dls_pgdn();
define dls_pgup();
define dls_position(name);
define dls_quit();
define dls_read_dir(dir);
define dls_refresh();
define dls_require();
define dls_save();
define dls_shape_list();
define dls_shellcmd();
define dls_test();
define dls_up();
define dls_upperdir();


%**********************************************************
%   Variables
%**********************************************************


private variable dls_cdir    = Null_String;
private variable dls_fmask   = Null_String;
private variable lMark       = NULL;
private variable nameCol     = 43;
variable InsertFromListing   = 0;


define dls_test ()
{

}


define dls_init_keymap ()
{
    reset_slim_keymap ();
    $1 = "dirlist_slimkmap";
    !if (keymap_p($1)) copy_keymap ($1, "slim_kmap");
    use_keymap ($1);

    %   jump-by-letter-press
    %   e.g. m to go the the next initial m
    %   allow upper ASCII characters
    variable i;
    for (i=32; i<256; i++)
        definekey ("dls_jump_to_initial", char(i), $1);

    undefinekey (Key_BS, $1);

    definekey ("dls_copy_all",     "^A",      $1);
    definekey ("dls_copy_line",    "^C",      $1);
    definekey ("dls_delete_file",  Key_Del,   $1);
    definekey ("dls_down",         Key_Down,  $1);
    definekey ("dls_end",          Key_End,   $1);
    definekey ("dls_grep",         "^F",      $1);
    definekey ("dls_help",         Key_F1,    $1);
    definekey ("dls_home",         Key_Home,  $1);
    definekey ("dls_menubar",      Key_F10,   $1);
    definekey ("dls_open_current", Key_Enter, $1);
    definekey ("dls_pgdn",         Key_PgDn,  $1);
    definekey ("dls_pgup",         Key_PgUp,  $1);
    definekey ("dls_quit",         "^W",      $1);
    definekey ("dls_refresh",      "^R",      $1);
    definekey ("dls_save",         "^S",      $1);
    definekey ("dls_shellcmd",     "^E",      $1);
    %definekey ("dls_test",         "^T",      $1);
    definekey ("dls_up",           Key_Up,    $1);
    definekey ("dls_upperdir",     Key_BS,    $1);
    definekey ("dtr",              Key_F6,    $1);

    undefinekey (Key_Right, $1);
    undefinekey (Key_Left,  $1);
}


define dls_jump_to_initial ()
%----------------------------------------------------------
%   Pressing a letter takes the cursor to the next entry
%   beginning with that letter. After the last entry with
%   that letter, the cursor jumps back to the first one,
%   i.e. jumping occurs in a ring. As an example,
%   pressing "a" when the cursor is resting on "t"
%   may take the cursor back in the listing.
%   Upper/lower case makes no difference.
%----------------------------------------------------------
{
    variable found=0;
    variable cf = 0;
    variable initl=strlow(LASTKEY);
    variable Initl=strup(LASTKEY);
    variable rply = Null_String;
    variable prmpt = "Not found. Cycle and restart from top (y/n)?";

    push_mark ();

    forever
    {
        eol ();
        if (eobp) break;
        found = fsearch ("  "+initl);
        !if(found) break;
        goto_column (nameCol);

        %   make this case insensitive
        if (looking_at(initl)) break;
        if (looking_at(Initl)) break;
        found=0;
    }

    if (found)
    {
        pop_mark(0);
        return;
    }

    %   nothing found below start position, cycle around

    rply = read_mini2 (prmpt, "n", "");

    !if (strlen(rply)) return;
    if (rply != "Y" and rply !="y") return;

    bob ();

    forever
    {
        eol ();
        if (eobp) break;
        found = fsearch ("  "+initl);
        !if(found) break;
        goto_column (nameCol);

        %   make this case insensitive
        if (looking_at(initl)) break;
        if (looking_at(Initl)) break;
        found=0;
    }

    if (found) pop_mark(0);
    else pop_mark(1);
}


define dls_quit ()
{
    sw2buf (DirListing);
    set_buffer_modified_flag (0);
    set_readonly (0);
    delbuf (DirListing);
}


define dls_load (name)
{
    %   will load a file for editing
    %   argument entails full path

    %   intrinsic read_file is find_file_cmd in ledit.c
    %   however there is a non-intrinsic read_file in file.c
    %   and it is called from find_file_cmd

    %   consider case file already opened in buffer
    %   while deleted on disk from external application
    %   you try to open it again from dirlisting
    %   it will just switch buffer with no error
    %   giving impression loading was ok but it was not

    !if (find_file (name))
    {
        delbuf(whatbuf);
        %   buffer opened to take file but file unavailable

        error ("Unable to find " + name +": Refresh listing!");

        %   Some files listed by ls return 0 (non existent),
        %   these are non-writable system files like /proc/kcore.
        %   Alert: non-writable system files can crash jed!
        return;
    }

    dls_quit ();

    %   c source is not tuned to sue's special buffers
    %   current buffer may not be where file was loaded
    %   so find it
    () = find_file (name);
}


define dls_octal_to_letters(o_string)
{
    %   o-string is the octal digit string
    %   defining permissions

    variable i, o, perms_letters=Null_String, edit=0, readf=0;
    variable type = substr (o_string,1,2);

    for (i=4;i<7;i++)
    {
        o = substr (o_string,i,1);

        switch (o)
        {
            case "0":
                perms_letters += "---";
        }
        {
            case "1":
                perms_letters += "--x";
        }
        {
            case "2":
                perms_letters += "-w-";
                edit += 1;
        }
        {
            case "3":
                perms_letters += "-wx";
                edit += 1;
        }
        {
            case "4":
                perms_letters += "r--";
                readf =+1;
        }
        {
            case "5":
                perms_letters += "r-x";
                readf =+1;
        }
        {
            case "6":
                perms_letters += "rw-";
                edit += 1;
                readf =+1;
        }
        {
            case "7":
                perms_letters += "rwx";
                edit += 1;
                readf =+1;
        }
    }

    %   leave out non-editable files
    %   some of them (like /proc/kmsg) would crash the editor

    if (type=="10" and edit==0 and readf==0)
        perms_letters = Null_String;
    %   allow only read, write: everything else out

    return perms_letters;
}


define dls_shape_list ()
{
    %   entries are already sorted by group
    %   in the order dir/regular/link
    %   sort by name in each group

    variable o_string=Null_String, perms_letters = Null_String;

    bob();

    if (bol_fsearch (" 4"))
    {
        %   directories
        newline();
        push_spot();

        %   to find the names, find the colon in hours
        %   then the separating two spaces, then skip
        %   nameCol not correct yet
        () = ffind (":");
        () = ffind ("  ");
        go_right(2);

        push_mark();
        eol();
        while (bol_fsearch (" 4")) eol();
        sort2 ();
        %   sort2 is the character stream sort

        %   now convert permissions from digits to letters
        pop_spot();

        while (bol_fsearch (" 4"))
        {
            push_mark();
            go_right(6);
            o_string = bufsubstr();
            perms_letters = dls_octal_to_letters(o_string);
            bol();
            loop (6) call ("delete_char_cmd");
            insert ("d"+perms_letters);
        }
    }

    if (bol_fsearch ("10"))
    {
        %   there are regular files
        newline();

        %   to find the names, find the colon in hours
        %   then the separating two spaces, then skip
        () = ffind (":");
        () = ffind ("  ");
        go_right(2);

        push_mark();
        eol();
        while (bol_fsearch ("10")) eol();
        sort2 ();
        %   sort2 is the character stream sort

        %   now convert permissions from digits to letters
        pop_spot();

        while (bol_fsearch ("10"))
        {
            push_mark();
            go_right(6);
            o_string = bufsubstr();
            perms_letters = dls_octal_to_letters(o_string);
            bol();

            %   if file non-editable skip line
            if (perms_letters==Null_String) delete_line ();

            else
            {
                loop (6) call ("delete_char_cmd");
                insert ("-"+perms_letters);
            }
        }
    }

    if (bol_fsearch ("12"))
    {
        %   symlinks
        newline();

        %   to find the names, find the colon in hours
        %   then the separating two spaces, then skip
        () = ffind (":");
        () = ffind ("  ");
        go_right(2);

        push_mark();
        eol();

        while (bol_fsearch ("12")) eol();
        sort2 ();
        %   sort2 is the character stream sort

        %   now convert permissions from digits to letters
        pop_spot();
        perms_letters = "l---------";

        while (bol_fsearch ("12"))
        {
            loop (6) call ("delete_char_cmd");
            insert (perms_letters);
        }
    }

    %   give the listing a header

    bob ();
    insert (dls_cdir);

    %   is the listing empty?
    dls_end();
    if (what_line() < 3)
    {
        go_down(1);
        newline();
        insert ("    No files, press Backspace");
    }
    else
    {
        bob();
        go_down(1);
    }

    set_buffer_modified_flag (0);
    set_readonly (1);
}


define dls_align_right(N,L)
{
    %   N,L are integer
    %   N is to be aligned right over L columns

    if (L<1) L=1;

    variable tmp = strlen (string(N));

    if (tmp > L)
    {
        loop(L) insert ("*");
        return;
    }

    variable cc = what_column ();
    goto_column(cc + (L-tmp));
    insert (string(N));
}


define dls_format_time (sstt)
{
    variable ttime  = localtime (sstt);
    variable yyear;
    variable mmonth = string(ttime.tm_mon + 1);
    variable dday   = string(ttime.tm_mday);
    variable hhours = string(ttime.tm_hour);
    variable mmins  = string(ttime.tm_min);

    if(ttime.tm_year > 0) yyear=string(ttime.tm_year+1900);
    else yyear="1900";

    if (strlen(dday)<2)   dday   = "0"+dday;
    if (strlen(mmonth)<2) mmonth = "0"+mmonth;
    if (strlen(hhours)<2) hhours = "0"+hhours;
    if (strlen(mmins)<2)  mmins  = "0"+mmins;

    %   formatted time needs 16 columns
    variable ftime  =
    yyear + "-" + mmonth + "-" + dday + " " + hhours + ":" + mmins;
    %   4    1         2    1       2    1         2    1        2

    return ftime;
}


define dls_draw_list (whichdir)
{
    ERROR_BLOCK _clear_error();

    variable A = listdir (whichdir);

    variable L = length(A);
    variable i=0, stf, stf2, type=0;
    variable dirslash="/", lnkslash=Null_String;

    bob();

    for (i=0;i<L;i++)
    {
        stf = lstat_file (A[i]);

        %   only insert details for dir, regular or link

        if(stat_is("lnk",stf.st_mode))      type=3;
        else if(stat_is("reg",stf.st_mode)) type=1;
        else if(stat_is("dir",stf.st_mode)) type=2;
        else continue;

        %   link may point to unsuitable object
        if (type==3)
        {
            stf2 = stat_file (A[i]);

            if (stf2 == NULL)
            {
                %   redraw screen to get rid
                %   of error messages from kernel
                call("redraw");
                continue;
            }

            lnkslash=Null_String;

            if(stat_is("lnk",stf2.st_mode)) continue;
            %   this is a link to a link, skip

            else if(stat_is("reg",stf2.st_mode));
            %   regular file, do nothing

            else if(stat_is("dir",stf2.st_mode)) lnkslash=dirslash;
            %   directory

            else continue;
            %   skip everything else

            %   links may remain that point to non-editable files
            %   such files will not be found and the link will
            %   give a message error: not a real nuisance
        }

        newline();
        %   convert mode to octal first
        %   then format on a length of 6
        dls_align_right(sprintf("%o",stf.st_mode),6);

        %   format size on a length of 12
        dls_align_right(stf.st_size,12);

        insert("  ");

        %   shorten and format time
        insert(dls_format_time(stf.st_mtime));

        insert("  ");

        %   insert file name (not link name)

        if (type==3) insert(readlink(A[i]));
        else insert(A[i]);

        if (type==2) insert(dirslash);
        else if (lnkslash==dirslash)
        {
            insert(dirslash);
            lnkslash=Null_String;
        }
    }

    %   jump back and sort by group
    %   i.e. dir/regular/link

    newline();
    push_mark();
    bob();

    sort2();
    call("delete_char_cmd");
}


define dls_read_dir (dir)
{
    %   parameter can be just a directory

    variable file, flags, s, tmpstr, old_dir;

    old_dir = dls_cdir;

    s = file_status (dir);

    if (s < 0)
    {
        flush ("Invalid entry.");
        dls_quit ();
        return;
    }

    if (s==2)
    {
        %   this is a directory, put it in canonical form

        tmpstr = dircat (dir, Null_String);
        %   dircat required for proper name without dots
        %   the second parameter is irrelevant

        dls_cdir = path_dirname (tmpstr);
        dls_fmask = Null_String;
    }

    %   allow exact hit to open immediately
    else if (s==1)
    {
        %   status 1: unique file exists

        (dls_cdir, dls_fmask) =
        (path_dirname (dir), path_basename (dir));

        tmpstr = dircat (dls_cdir, dls_fmask);
        dls_load (tmpstr);
        return;
    }

    else
    {
        %   status 0: file mask

        (dls_cdir, dls_fmask) =
        (path_dirname (dir), path_basename (dir));
    }

    sw2buf (DirListing);

    (file,,,flags) = getbuf_info ();

    setbuf_info (file, dls_cdir, DirListing, flags);

    set_status_line
    ("  " + dls_cdir + "  |  %p  |  F1=Help  F10=Menu  ", 0);

    dls_init_keymap ();
    set_readonly (0);
    erase_buffer ();

    ()  = chdir (dls_cdir);

    dls_draw_list (dls_cdir);

    %   list is sorted by dir/regular/link
    %   now sort by name within each group, case-insensitive

    dls_shape_list ();
}


define dls_refresh ()
{
    call("redraw");
    dls_read_dir (dls_cdir);
}


define dls_down ()
{
    go_down (1);
    bol ();
}


define dls_up ()
{
    go_up (1);
    bol ();
}


define dls_home ()
{
    %if (bolp) bob ();
    %else bol ();
    bob();
}


define dls_end ()
{
    eob();
    while(is_whiteline) dls_up();
}


define dls_getfile ()
{
    variable name, type, tmps;

    % (name, type) = dls_getfile ();
    % name = name of file or directory

    % type = -1 : other
    % type = 0  : file read-only
    % type = 1  : file editable
    % type = 2  : directory
    % type = 3  : link

    bol ();
    if (looking_at ("l"))      type = 3;
    else if (looking_at ("d")) type = 2;
    else if (looking_at ("-"))
    {
        push_mark();
        go_right(10);
        tmps=bufsubstr();
        type = is_substr (tmps,"w");
        if (type) type=1;
    }
    else return (Null_String, -1);

    goto_column (nameCol);
    push_mark ();
    eol ();
    name = bufsubstr ();

    return (name, type);
}


define dls_menubar ()
{
    select_pulldown_file();
}


define dls_help ()
{
    select_pulldown_help ();
}


define dls_display_help ()
{
    variable
        stringA="DirListing",
        stringB="press F1",
        stringC="&Quick Help: Dir Listing";

    display_hlp2 (stringA,stringB,stringC);
}


define dls_create_menu ()
{
    %   The menu bar is created only once.
    %   When the create command is issued again,
    %   there is a check for the existence of the menu bar.
    %   To avoid garbage, delete before appending.

    variable tmpbuf = whatbuf ();

    sw2buf (DirListing);

    menu_create_menu_bar (DirListing);

    menu_use_menu_bar (DirListing);
    menu_append_popup (DirListing, "&File");
    menu_append_popup (DirListing, "&Tools");
    menu_append_popup (DirListing, "&Help");

    $1 = DirListing+".&File";

    menu_delete_items ($1);
    menu_append_item ($1, "&Open", "dls_open_current");
    menu_append_popup ($1, "Open &With");
    menu_append_item ($1, "&Insert", "dls_insert_from_listing");
    menu_append_item ($1, "&Delete", "dls_delete_file");
    menu_append_separator ($1);
    menu_append_item ($1, "&Close Listing", "dls_quit");
    menu_append_item ($1, "&Save Listing", "dls_save");
    menu_append_item ($1, "&Refresh Listing", "dls_refresh");
    menu_append_separator ($1);
    menu_append_item ($1, "&Upper Directory", "dls_upperdir");
    menu_append_item ($1, "&Tree", "dtr");
    menu_append_separator ($1);
    menu_append_item ($1, "E&xit", "disk_exit");

    $1 = DirListing+".&File.Open &With";

    menu_delete_items ($1);
    menu_append_item ($1, "&Lynx", "dls_lynx");
    menu_append_item ($1, "&Man", "dls_man");

    $1 = DirListing+".&Tools";

    menu_delete_items ($1);
    menu_append_item ($1, "Copy &Line", "dls_copy_line");
    menu_append_item ($1, "Copy &All", "dls_copy_all");
    menu_append_separator ($1);
    menu_append_item ($1, "&Find in Files", "dls_grep");
    menu_append_item ($1, "&Shell Command", "dls_shellcmd");
    menu_append_item ($1, "A&dd Functions", "dls_require");

    $1 = DirListing+".&Help";

    menu_delete_items ($1);
    menu_append_item ($1, "&Quick Help", "dls_display_help");
    menu_append_item ($1, "&List Keys", "dls_list_keys");

    sw2buf (tmpbuf);
}


define dls (dir)
{
    %   entry point from outside, starts the listing

    %   store buffer name for possible insertion of files
    !if (is_substr(whatbuf(),"/"))
    {
        if(InsBuffer==Null_String)
            (InsBuffer,,,) = getbuf_info ();
    }
    else InsBuffer=Null_String;

    sw2buf (DirListing);
    dls_init_keymap ();
    dls_create_menu ();
    dls_read_dir (dir);
}


define dls_position (name)
%   used for links to files
%   link is not opened, cursor is positioned on it
%   position may not be exact due to case
{
    variable found;
    variable p_name = path_dirname (name);
    variable f_name = path_basename (name);

    dls_read_dir (p_name);
    bob ();

    forever
    {
        found = fsearch (f_name);

        !if (found)
        {
            if(strlen(f_name))
                flush ("Not found: "+f_name+" !");
            bob ();
            break;
        }

        if (nameCol == what_column ()) break;
        eol ();
    }
}


define dls_open_current ()
%----------------------------------------------------------
%   Shift not pressed:  will open a file from DirListing
%                       in own buffer
%
%   Shift pressed:      will insert file from DirListing
%                       into current buffer
%----------------------------------------------------------
{
    %   called when pressing enter on a name in DirListing

    if (SHIFTKEY_PRESSED) InsertFromListing=1;
    else InsertFromListing=0;

    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    variable name, type;

    (name, type) = dls_getfile ();

    name = dircat (dls_cdir, name);

    if (InsertFromListing and (type==1 or type==0))
    {
        dls_insert_current ();
        InsertFromListing=0;
        return;
    }

    if (type==0)
    {
        %   this is a file read-only
        flush("File not for editing: "+ name);
        return;
    }

    if (type==1)
    {
        %   this is a file for editing

        LASTDIR = dls_cdir;
        dls_load (name);
        %   loading wil close dir listing
    }

    else if (type==3) dls_position (name);
    %   only position cursor on reference, do not open

    %   this should be a directory or else trouble
    else
    {
        flush ("Reading directory ...");
        dls_read_dir (name);
        flush (" ");
    }
}


define dls_insert_from_listing ()
{
    InsertFromListing=1;
    dls_open_current();
    InsertFromListing=0;
}


define dls_upperdir ()
{
    variable old_dir = dls_cdir;

    () = chdir ("..");

    dls_read_dir (".");

    dls_position (old_dir);
}


define dls_pgup ()
{
    variable start_line = window_line ();
    %   cursor is on start_line

    goto_top_of_window ();
    % move to top of window, if not already there

    if (window_line () != start_line) return;
    %   really moved, current line is not start_line

    else
    {
        %   already top, now paginate
        call ("page_up");
        goto_top_of_window ();
    }
}


define dls_pgdn ()
{
    variable start_line = window_line ();
        % cursor is on start_line

    goto_bottom_of_window ();
        % move to bottom of window, if not already there

    if (window_line () != start_line) return;
        %   really moved, current line is not start_line

    else
    {
        % already at bottom of window, now paginate
        call ("page_down");
        goto_bottom_of_window ();
    }
}


define dls_insert_current ()
{
    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    if (InsBuffer==Null_String)
    {
        verror ("Don't know where to insert.");
        dls_quit ();
    }

    %   buffer for insertion is known
    %   this function called ONLY from dls_open_current

    variable name, type;

    (name, type) = dls_getfile ();

    name = dircat (dls_cdir, name);

    %   insertion buffer was set when dls called up

    if (type==1 or type==0)
    {
        % file is editable or read-only
        sw2buf (InsBuffer);
        !if (down(1))
        {
            eol();
            insert_char ('\n');
        }
        newline();
        %   necessary for rare cases of empty file

        insert_file (name);
        dls_quit ();
        return;
    }

    verror ("Unable to insert!");

    dls_quit ();
}


define dls_save ()
{
    variable b_name = Null_String;

    b_name = read_path_from_mini ("Save listing to file:");

    %   if response empty or aborted, do nothing
    if (b_name == "") return;

    %   aborted without pressing enter
    if (LASTKEY[0] != 13) return;

    %   select entire listing for saving

    push_spot ();
    bob ();
    push_mark ();
    eob ();

    %   check file status before proceeding

    variable b_status = file_status (b_name);

    if (b_status == 2)
        error ("Name already in use for a directory");

    else if (b_status == -1)
        error ("File exists and is not accessible");

    else if (b_status == -2)
        error ("Invalid path");

    else if (b_status == -3)
        error ("Undetermined error");

    else if (b_status == 0)
    {
        write_region_to_file (b_name);
    }

    else if (b_status == 1)
    {
        variable ch = get_mini_response
        ("File exists: append/overwrite/cancel? (a/o/c) ");

        if (ch=='a' or ch=='A')
            append_region_to_file (b_name);

        else if (ch=='o' or ch=='O')
            write_region_to_file (b_name);
    }
    pop_spot ();
    dls_refresh (dls_cdir);
}


define dls_copy_all ()
{
    %   copy entire directory listing to clipboard

    push_spot ();
    bob ();
    push_mark ();
    eob ();
    call ("copy_region");
    pop_spot ();
    flush ("Copied all to clipboard.");
}


define dls_copy_line ()
{
    %   copy current line to clipboard

    push_spot ();
    bol ();
    push_mark ();
    eol ();
    call ("copy_region");
    pop_spot ();
    flush ("Current line copied to clipboard.");
}


define dls_delete_file ()
{
    %   called when pressing Del on a name in DirListing

    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    variable name, type;

    (name, type) = dls_getfile ();

    %   take note of following file to position cursor after deleting
    go_down(1);
    variable name_nxt=Null_String;
    !if (is_whiteline) (name_nxt,) = dls_getfile ();
    go_up(1);

    name = dircat (dls_cdir, name);

    if (type==1 or type==0)
    {
        %   this is a file for editing

        variable ch = get_mini_response
        ("Delete current file? (y/n)  ");

        if (ch!='y' and ch!='Y') return;

        %   run remove command suppressing error messages
        () = run_shell_cmd
        ("rm  -f \"" + name + "\" &> /dev/null");

        dls_refresh ();
        if(strlen(name_nxt))
            ()=fsearch("  "+name_nxt);
        else eob();
        while (is_whiteline ()) go_up (1);
    }

    else flush ("No.");
}


define dls_lynx ()
{
    %   called from menu in DirListing

    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    variable name, type;

    (name, type) = dls_getfile ();

    name = dircat (dls_cdir, name);

    %   better (a little) safe than sorry
    !if (is_substr(name,".htm"))
    {
        flush ("No.");
        return;
    }

    if (type==1 or type==0)
    {
        LASTDIR = dls_cdir;

        variable tmpbf = string(_time);
        variable cmdstr =
        "lynx -dump -force_html " + name +"> " + tmpbf;
        %   Let lynx render one single html page and exit.
        %   Redirect page to text file
        %   to be read (and deleted) in a blink.

        if (0 != run_program (cmdstr)) return;

        dls_quit();

        sw2buf(name);
        insert_file (tmpbf);
        () = run_shell_cmd ("rm " + tmpbf);

        bob();
        set_readonly(1);

        set_status_line (" lynxview: " + name + "  %p  F1=Help  ", 0);

        %   the html file is now in a read-only buffer
        %   the intermediate file on disk is deleted
    }
}


define dls_man ()
%----------------------------------------------------------
%   open a file as man page
%   the file MUST be a man page or it's going to be garbage
%----------------------------------------------------------
{
    if (bolp and eolp) return;
    %   nothing to do on empty lines

    if (what_line () < 3 ) return;
    %   nothing to do at the top

    variable name, type;

    (name, type) = dls_getfile ();

    if (type!=1 and type!=0) return;

    name = dircat (dls_cdir, name);

    %   better (a little) safe than sorry
    variable i, suitable_file=0;
    if (is_substr(name,".n.")) suitable_file=1;
    for (i=1;i<10;i++)
        if (is_substr(name,"."+string(i))) suitable_file=1;

    if (suitable_file==0)
    {
        flush ("No.");
        return;
    }

    %   by now, file is likely to be manpage


    LASTDIR = dls_cdir;
    dls_quit ();

    variable buf = "man page" +" "+ name;
    %   useless for the moment being

    sw2buf (buf);
    erase_buffer ();

    variable rsc =
    () = run_shell_cmd (sprintf ("man %s 2> /dev/null", name));

    if (rsc==0)
    {
        display_clean_manpage ();
        bob ();
        set_buffer_modified_flag (0);
    }

    else delbuf (buf);
}


define dls_grep ()
{
    variable
        sstr    = Null_String,
        fmask   = Null_String,
        fopts   = Null_String,
        cmdstr  = Null_String,
        result  = Null_String,
        rfile   = "grep_results.tmp";

    sstr = read_mini2 ("Find what:", Null_String, Null_String);

    !if (strlen(sstr)) return;
    if (LASTKEY[0] == 23) return;

    %   make sure search string is quoted
    !if (substr (sstr,1,1)=="\"")
        sstr = "\""+ sstr + "\"";

    fmask = read_mini2 ("Files to grep:", "*", "*");

    !if (strlen(strtrim(fmask))) return;
    if (LASTKEY[0] == 23) return;

    fopts = read_mini2 ("Options to use:", "in", "in");
    if (LASTKEY[0] == 23) return;

    fopts = "-s" + fopts;

    () = delete_file (rfile);
    sw2buf(rfile);
    erase_buffer();

    %cmdstr = "grep -ins " + sstr + " " + fmask + " 2> /dev/nul &";
    cmdstr = "grep " + fopts + " " + sstr + " " + fmask + " 2> /dev/nul &";

    flush ("Grepping ... To stop press Control_Backslash.");

    result = run_shell_cmd (cmdstr);

    %   as long as grep is running,
    %   control will be in the minibuffer
    %   actually in a separate process

    !if (MINIBUFFER_ACTIVE)
    {
        if (result>0)
        {
            flush("Not found. Executed: " + cmdstr);
            delbuf (rfile);
            dls (".");
            return;
        }

        else
        {
            %   save buffer to disk
            () = write_buffer (rfile);
            delbuf (rfile);
            dls_refresh ();
            fsearch(rfile);
            flush("Executed: " + cmdstr);
            return;
        }
    }

    %   there might be no files corresponding to mask
    cmdstr = "ls " + fmask + " 2> /dev/nul";
    if (run_shell_cmd (cmdstr))
    {
        flush ("No files match specifications entered.");
        return;
    }

    if (result==2)
    {
        dls_refresh ();
        reset_screen();
        flush("Error in input: retry using quotes.");
        %dls (".");
    }
}


define dls_list_keys ()
{
    set_readonly (0);
    erase_buffer ();
    bob ();

    dump_bindings ("dirlist_slimkmap");

    %   suppress lines containing self-insert characters
    bob ();
    while (fsearch ("dls_jump_to_initial")) delete_line ();

    readable_keys();

    bob ();
    replace     ("dtr", "Tree View");

    bob ();
    replace     ("^M   ",  "Enter");

    bob ();
    push_mark ();
    eob();
    go_left(1);
    sort2 ();
    while (is_whiteline) call("delete_char_cmd");
    %   this sort is not case sensitive

    set_readonly(1);
    bob ();
    set_buffer_modified_flag (0);
    set_status_line("  Press any key to exit  ",0);
    variable m = "&Keybindings DirListing";
    dummy_menu(m);
    update(0);

    flush_input(getkey);
    delbuf(whatbuf);
}


define dls_shellcmd ()
%----------------------------------------------------------
%   Execute a non-interactive shell command from DirListing
%   keeping cursor approx. on same screen line
%   sending output to a file in current directory
%----------------------------------------------------------
{
    ERROR_BLOCK
    {
        _clear_error ();
        reset_screen();
    }

    variable
        sstr    = Null_String,
        result  = -1,
        rfile   = "shellcmd_results.tmp",
        type    = 0,
        name    = Null_String;

    (name, type) = dls_getfile ();
    if (type==2) name = dircat (dls_cdir, name);

    sstr = read_mini2 ("*Non-interactive* Shell Command:", name, name);
    if (LASTKEY[0] == 23) return;
    !if (strlen(sstr)) return;

    () = delete_file (rfile);
    sw2buf(rfile);
    erase_buffer();

    result = run_shell_cmd (sstr);

    %   as long as the shell command is running,
    %   control will be in the minibuffer

    !if (MINIBUFFER_ACTIVE)
    {
        if (result > 0)
        {
            reset_screen();
            flush("Errors! Command: " + sstr);
            delbuf (rfile);
        }

        else
        {
            %   save buffer to disk
            () = write_buffer (rfile);
            delbuf (rfile);
            dls_refresh ();
            fsearch(rfile);
            flush("Executed: " + sstr);
        }
    }
}


define dls_require ()
%----------------------------------------------------------
%   Add an SL function module - named according to the pattern
%   *.sl - to the namespace AddSpace. Functions so added
%   can then be executed via the menu in an editing buffer.
%----------------------------------------------------------
{
    if (strcmp(strcat (JED_ROOT,"/lib/"), dls_cdir) <0);
    else
    {
        flush("Please issue command from below lib directory");
        return;
    }

    ERROR_BLOCK
    {
        _clear_error();
        flush("Failure: 'require' failed.");
        return;
    }

    variable type = 0;
    variable name = Null_String;
    variable xtnsion = Null_String;
    variable reqst = 1;

    (name, type) = dls_getfile ();

    !if(strlen(strtrim(name))) reqst=0;
    if (type!=1 and type!=0) reqst=0;

    if (reqst)
    {
        xtnsion =  substr(name,strlen(name)-2,3);
        if(xtnsion != ".sl") reqst=0;
    }

    if (reqst)
    {
        () = evalfile (name, AddSpace);
        flush("OK.");
    }
    else flush ("No.");
}
